"""
Compare screenshots taken via the webdriver.
"""

# Copyright (c) 2013 contributors; see AUTHORS.
# Licensed under the Apache License, Version 2.0
# https://www.apache.org/licenses/LICENSE-2.0

import math

try:
    # Pillow
    from PIL import Image
    from PIL import ImageChops
except ImportError: # pragma: no cover
    # PIL
    try:
        import Image # pylint: disable=F0401
        import ImageChops # pylint: disable=F0401
    except ImportError:
        raise ImportError('Could not import Pillow or PIL')

from gossamer import util, exc

def allowance(browser):
    """
    Our image diffs below give some false alarms of diffs. For the browser
    Chrome, which takes screenshots limited to screensize, a value of at least
    573 is always seen. For Firefox, which takes screenshots the width of
    the actual screensize and the height of the document, 958.
    """
    margins = {'default': 573, 'chrome': 573, 'firefox': 958}
    try:
        return margins[browser]
    except KeyError:
        return margins['default']


def images_identical(path1, path2, margin=None):
    """
    Hacky test of images being identical. PIL can show incorrect diffs.
    """
    util.log.debug('images_identical: %s, %s', path1, path2)
    margin = margin or 0
    im1 = Image.open(path1)
    im2 = Image.open(path2)
    rmsdiff = _rmsdiff_2011(im1, im2)
    if rmsdiff <= margin:
        return True
    else:
        return False


def image_diff(path1, path2, outpath, diffcolor):
    """
    Generate a diff image on a screenshot which has failed
    :func:`.images_identical`.
    """
    im1 = Image.open(path1)
    im2 = Image.open(path2)

    rmsdiff = _rmsdiff_2011(im1, im2)

    pix1 = im1.load()
    pix2 = im2.load()

    if im1.mode != im2.mode:
        raise exc.TestError(
            'Different pixel modes between %r and %r' % \
            (path1, path2)
        )
    if im1.size != im2.size:
        raise exc.TestError(
            'Different dimensions between %r (%r) and %r (%r)' % \
            (path1, im1.size, path2, im2.size)
        )

    mode = im1.mode

    if mode == '1':
        value = 255
    elif mode == 'L':
        value = 255
    elif mode == 'RGB':
        value = diffcolor
    elif mode == 'RGBA':
        value = diffcolor + (255,)
    elif mode == 'P':
        raise NotImplementedError('Need to look up nearest palette color')
    else:
        raise NotImplementedError('Unexpected PNG mode')

    width, height = im1.size

    for y in xrange(height):
        for x in xrange(width):
            if pix1[x, y] != pix2[x, y]:
                pix2[x, y] = value
    im2.save(outpath)

    return (rmsdiff, width, height)


def _rmsdiff_2011(im1, im2):
    "Calculate the root-mean-square difference between two images"
    diff = ImageChops.difference(im1, im2)
    h = diff.histogram()
    sq = (value * (idx ** 2) for idx, value in enumerate(h))
    sum_of_squares = sum(sq)
    rms = math.sqrt(sum_of_squares / float(im1.size[0] * im1.size[1]))
    util.log.debug('rmsdiff: %s', rms)
    return rms



"""
Command-line interface to the app.
"""

# Copyright (c) 2013 contributors; see AUTHORS.
# Licensed under the Apache License, Version 2.0
# https://www.apache.org/licenses/LICENSE-2.0

import glob
import os
import sys

import plac # pylint: disable=F0401

from gossamer.main import dispatch
from gossamer.constant import modes, exits, states, \
    DEFAULT_WEBDRIVER, DEFAULT_TESTFILE, \
    DEFAULT_DIFFCOLOR, DEFAULT_SCREENSIZE, \
    DEFAULT_BROWSER
from gossamer import util, exc
from gossamer import __version__


@plac.annotations(
    names = plac.Annotation(
        'Test case name(s) to use, comma-separated',
    ),
    postdata = plac.Annotation(
        'File for POST data or - for stdin'
    ),

    testfile = plac.Annotation(
        'Test file(s) to use',
        'option', 'f', str,
        metavar='GLOB'
    ),
    record = plac.Annotation(
        'Record a new test',
        'flag', 'r'
    ),
    rerecord = plac.Annotation(
        'Re-run the test but take new screenshots',
        'flag', 'rr'
    ),

    local = plac.Annotation(
        'Local WebDriver URL to use',
        'option', 'l',
        metavar=DEFAULT_WEBDRIVER
    ),
    remote = plac.Annotation(
        'Remote WebDriver to use',
        'option', 'w',
        metavar=DEFAULT_WEBDRIVER
    ),

    browser = plac.Annotation(
        'Browser to use, either firefox, chrome, phantomjs, ie, or opera',
        'option', 'b', str,
        metavar=DEFAULT_BROWSER,
    ),
    screensize = plac.Annotation(
        'Width and height for screen (i.e. 1024x768)',
        'option', 'z',
        metavar=DEFAULT_SCREENSIZE
    ),
    diffcolor = plac.Annotation(
        'Diff color for errors in RGB (i.e. 0,255,0)',
        'option', 'p', str,
        metavar=DEFAULT_DIFFCOLOR
    ),

    save_diff = plac.Annotation(
        'Save information about failures as last.png and diff.png',
        'flag', 'e'
    ),

    overwrite = plac.Annotation(
        'Overwrite existing tests without asking',
        'flag', 'o'
    ),

    data_dir = plac.Annotation(
        'Directory in which tests should be stored',
        'option', 'd', str,
    ),

    version = plac.Annotation(
        'Get the current version',
        'flag', 'version'
    ),

    verbose = plac.Annotation(
        'Verbosity, with -v as logging.DEBUG',
        'flag', 'v', 'verbose'
    ) # pylint: disable=R0915,R0912,R0911,R0914
)


def initialize(
        names=None,
        testfile=None,
        record=False,
        rerecord=False,
        local=None,
        remote=None,
        postdata=None,
        browser=None,
        screensize=None,
        diffcolor=None,
        save_diff=False,
        overwrite=False,
        data_dir=None,
        version=False,
        verbose=False
    ): # pylint: disable=R0913,W0613
    """
    Gossamer CLI.
    """

    if version:
        sys.stdout.write('Gossamer %s\n' % __version__)
        sys.stdout.flush()
        return exits.OK

    sys.stdout.write('Initializing gossamer and opening WebDriver...\n')
    sys.stdout.flush()

    if verbose:
        util.log = util.logger(__name__, 'DEBUG')

    names = names.split(',') if names else None
    browser = browser or DEFAULT_BROWSER

    cwd = os.getcwd()
    test_files = []
    for pattern in (testfile or DEFAULT_TESTFILE).split(','):
        for name in glob.glob(pattern):
            test_files.append(os.path.join(cwd, name))
    if len(test_files) == 0:
        sys.stdout.write('No Gossamerfile found.\n')
        sys.stdout.flush()
        return exits.ERROR

    # data_dir
    if not data_dir:
        data_dir = os.path.join(cwd, 'gossamer')
        sys.stdout.write('Default data directory of %s\n' % data_dir)
        sys.stdout.flush()
    else:
        if not os.path.isabs(data_dir):
            data_dir = os.path.join(os.getcwd(), data_dir)

    # mode
    if record and rerecord:
        sys.stdout.write('Cannot specify both -r and -rr\n')
        sys.stdout.flush()
        return exits.ARGUMENT_ERROR
    if record:
        mode = modes.RECORD
    elif rerecord:
        mode = modes.RERECORD
    else:
        mode = modes.PLAYBACK

    attrs = (
        'names', 'local', 'remote', 'postdata',
        'browser', 'screensize', 'diffcolor', 'save_diff', 'overwrite'
    )
    options = {
        key: val for key, val in \
        [(each, locals()[each]) for each in attrs]
    }

    # make tests using the test_files and mode we've resolved to
    try:
        tests = util.make_tests(test_files, mode, data_dir, **options)
    except exc.DoNotOverwrite as exception:
        sys.stdout.write(str(exception))
        sys.stdout.write('\n')
        sys.stdout.flush()
        return exits.ERROR
    except (exc.RecordedRunDoesNotExist, exc.RecordedRunEmpty,
        exc.CouldNotParseRecordedRun) as exception:
        sys.stdout.write(str(exception))
        sys.stdout.write('\n')
        sys.stdout.flush()
        return exits.RECORDED_RUN_ERROR

    # let the user mix browsers within tests by switching if browser has changed
    # this is really ugly, because main.dispatch assumes a batch of tests
    results = {}
    errs = {}

    driver = None
    try:
        for key, test in tests.items():
            if driver is not None:
                util.close_driver(driver)
            driver = util.get_driver(test.settings.browser, local, remote)
            if not driver:
                raise exc.WebDriverConnectionFailed(
                'We cannot connect to the WebDriver %s -- is it running?' % local
            )
            # run the tests
            try:
                result, err = dispatch(driver, mode, {key: test})
                results.update(result)
                errs.update(err)
            except exc.NoScreenshotsRecorded as exception:
                sys.stdout.write(str(exception))
                sys.stdout.flush()
                return exits.ERROR
            sys.stdout.write('\n')
            sys.stdout.flush()
    finally:
        util.close_driver(driver)

    if mode == modes.PLAYBACK:
        fails = sum(x is states.FAIL for _, x in results.items())
        errors = sum(x is states.ERROR for _, x in results.items())
        if fails > 0 or errors > 0:
            msg = []
            if fails > 0:
                msg.append('failed=%s' % fails)
            if errors > 0:
                msg.append('errors=%s' % errors)
            sys.stdout.write(
                'FAILED (%s)\n' % ', '.join(msg)
            )
            sys.stdout.flush()
            return exits.FAILED
        else:
            sys.stdout.write('OK\n')
            sys.stdout.flush()
            return exits.OK
    return exits.OK


def main():
    """
    Defined as the `gossamer` command in setup.py.

    Runs the argument parser and passes settings to
    :func:`gossamer.main.dispatcher`.
    """
    try:
        sys.exit(plac.call(initialize))
    except KeyboardInterrupt:
        sys.stdout.write('\n')
        sys.stdout.flush()
        sys.exit(1)



"""
Data structures that do not represent user actions. For those, see `step`.
"""

# Copyright (c) 2013 contributors; see AUTHORS.
# Licensed under the Apache License, Version 2.0
# https://www.apache.org/licenses/LICENSE-2.0


class Test(object): # pylint: disable=R0903
    """
    Object to be passed into dispatch... containing all information
    to run (and repeat, if persisted) a test.
    """

    def __init__(self, version, settings, steps):
        self.version = version
        self.settings = settings
        self.steps = steps

    def __json__(self):
        return {
            self.settings.name: {
                'version': self.version,
                'settings': self.settings,
                'steps': self.steps,
            }
        }

    def __repr__(self): # pragma: no cover
        return "<%s %r>" % (
            self.__class__.__name__,
            self.settings.name
        )


class Settings(object): # pylint: disable=R0903,R0902
    """
    Hold validated settings for a specific test run.
    """

    def __init__(self,
            name, url, mode, path, browser,
            screensize, postdata,
            diffcolor, save_diff, cookies=None, desc=None
        ): # pylint: disable=R0913
        self.name = name
        self.url = url
        self.mode = mode
        self.path = path
        self.browser = browser
        self.screensize = screensize
        self.postdata = postdata
        self.diffcolor = diffcolor
        self.save_diff = save_diff
        self.desc = desc
        self.cookies = cookies
        if self.cookies:
            self._validate_cookies()

    def navigate(self):
        """
        Return data in form expected by :func:`gossamer.run.navigate`.
        """
        return (self.url, self.postdata)

    def _validate_cookies(self):
        """
        Validate cookies for Selenium.
        """
        # http://selenium-python.readthedocs.org/en/latest/api.html#module-selenium.webdriver.remote.webdriver
        if not iter(self.cookies):
            raise ValueError('Cookies must be given as an iterable of dictionaries')
        if len(set([cookie['domain'] for cookie in self.cookies])) > 1:
            raise ValueError('Can only specify cookies for one domain')
        for cookie in self.cookies:
            if not isinstance(cookie, dict):
                raise ValueError('Needed a dictionary for cookie')
            for attr in ('name', 'domain', 'value'):
                if not attr in cookie:
                    raise ValueError('Cookie missing required attribute %s' % attr)

    def __json__(self):
        return self.__dict__

    def __repr__(self): # pragma: no cover
        return '<%s %r>' % (self.__class__.__name__, self.__dict__)


class Point(object): # pylint: disable=R0903
    """
    Contains validated x, y coordinates for screen position.
    """

    def __init__(self, x, y):
        """
        Stores x and y coordinates. They cannot be negative.
        """
        if x < 0 or y < 0: # pragma: no cover
            raise ValueError(
                'Coordinates [%s, %s] cannot be negative' % (x, y)
            )
        self.x = x
        self.y = y

    def __json__(self):
        return self.__dict__

    def __repr__(self): # pragma: no cover
        return '<Point %s, %s>' % (self.x, self.y)



"""
Exceptions
"""

# Copyright (c) 2013 contributors; see AUTHORS.
# Licensed under the Apache License, Version 2.0
# https://www.apache.org/licenses/LICENSE-2.0

class TestError(RuntimeError):
    """
    Generic.
    """

class NoScreenshotsRecorded(Exception):
    """
    A test is recorded with no screenshots made.
    """

class RecordedRunEmpty(Exception):
    """
    A record.json file is empty, meaning a run didn't complete.
    """

class DuplicateTestName(Exception):
    """
    A test name is specified twice.
    """

class CouldNotParseRecordedRun(Exception):
    """
    Couldn't read a record.json for some reason.
    """

class RecordedRunDoesNotExist(Exception):
    """
    A record.json file does not exist.
    """

class UnknownTestName(Exception):
    """
    Test name given via CLI is not known from a file.
    """

class DoNotOverwrite(Exception):
    """
    User indicates they don't want to overwrite an existing test.
    """

class EmptyPath(Exception):
    """
    In playing back, a data directory did not exist.
    """

class DifferentBrowser(Exception):
    """
    The recorded browser is different than the user-supplied browser.
    """

class InvalidBrowser(Exception):
    """
    We don't recognize a given `browser` value.
    """

class InvalidGossamerfile(Exception):
    """
    An error parsing the Gossamerfile.
    """

class InvalidWebDriverConfiguration(Exception):
    """
    WebDriver throws an exception while setting up.
    """

class ImageNotFound(Exception):
    """
    Imagediff was passed a filename to an image that doesn't exist.
    """

class ScreenshotsDiffer(Exception):
    """
    Screenshots are different... failed.
    """

class PlaybackTimeout(Exception):
    """
    We waited for the page to be unchanging (via watching mutations),
    and the wait exceeded our timeout.
    """

class ElementNotVisible(Exception):
    """
    WebDriver reports element not visible... not enough delay when something
    was being hidden/unhidden or rendered.
    """

class UnavailableBrowser(Exception):
    """
    WebDriver isn't starting with the user's browser selected.
    """

class WebDriverConnectionFailed(Exception):
    """
    Cannot connect to the webdriver--is it running?
    """

class WebDriverWentAway(Exception):
    """
    WebDriver raises an inability to communicate with the user agent.
    """

class WebDriverSecurityError(Exception):
    """
    WebDriver gave a security error.
    """



"""
See main.py.
"""

# Copyright (c) 2013 contributors; see AUTHORS.
# Licensed under the Apache License, Version 2.0
# https://www.apache.org/licenses/LICENSE-2.0

from gossamer.integration import run_gossamerfile

__all__ = ['run_gossamerfile', ]

__version__ = '0.9.3'




"""
Integrate with unittest.
"""

# Copyright (c) 2013 contributors; see AUTHORS.
# Licensed under the Apache License, Version 2.0
# https://www.apache.org/licenses/LICENSE-2.0

import unittest

from gossamer.main import dispatch
from gossamer.constant import modes, states, LOCAL_WEBDRIVER_URL
from gossamer import util, exc


def run_gossamerfile(
        client_locals, gossamerfile, data_dir,
        selenium=None, skip_allowed=True, rewrite_url=None
    ): # pylint: disable=R0913
    """
    Call this to read one or more Gossamerfiles and run all of their tests.
    It will mutate the locals() passed to it by the inclusion of a complete
    `unittest.TestCase` for every test in the given Gossamerfiles. Test
    runners will then automatically add your tests.

    Parameters:

        client_locals, locals() dictionary:
            `locals()` of the module from which the func is being called.

        gossamerfile, {str, list, tuple, dict}:
            Location to one or more Gossamerfiles.

        data_dir:
            The data directory containing the recorded tests and screenshots.

        selenium (optional), str:
            If provided, the Selenium Server URL to use instead of that in
            the recorded tests. Use this to change servers with environments.

        skip_allowed (optional), bool:
            If true, if Selenium Server is not running,
            unittest will skip this test; if false, it will error. Default
            true.

        rewrite_url (optional), callable:
            If given, test URLs will be rewritten according to the provided
            callable. This callable should accept a single parameter, a string,
            which is the URL in the recorded test. Use this to change the
            environment used. E.g., lambda x: x.replace('http://dev.', 'http://ci.').

    """
    if isinstance(gossamerfile, (str, unicode)):
        gossamerfile = [gossamerfile]
    elif isinstance(gossamerfile, dict):
        gossamerfile = [val for _, val in gossamerfile.items()]

    selenium = selenium or LOCAL_WEBDRIVER_URL
    options = {'local': selenium, 'remote': selenium}

    driver_ok = util.check_driver(selenium)

    tests = util.make_tests(
        gossamerfile, modes.PLAYBACK, data_dir, rewrite_url=rewrite_url, **options
    )
    for key, test in tests.items():
        case = type(
            'GossamerTestCase',
            GossamerTestCase.__bases__,
            dict(GossamerTestCase.__dict__)
        )
        case._skip_allowed = skip_allowed # pylint: disable=W0212
        case._driver_ok = driver_ok # pylint: disable=W0212
        case.runTest = _nose_wrapper(
            case, key, test, test.settings.browser, selenium, selenium
        )
        case._gossamer_test = test # pylint: disable=W0212
        case.runTest.__func__.__doc__ = test.settings.desc or test.settings.name # pylint: disable=E1101,C0301
        client_locals['GossamerTest_%s' % key] = case
    return True


def _nose_wrapper(self, *args): # pylint: disable=W0613
    """
    Creates GossamerTestCase methods that will themselves run the tests.
    """
    return lambda x: _run_nose(x, *args)


def _run_nose(self, name, test, browser, local, remote): # pylint: disable=R0913
    """
    Run a test for Nose.
    """
    try:
        driver = util.get_driver(browser, local, remote)
        log = dispatch(
            driver, modes.PLAYBACK, {name: test, }, output=util.null_writer
        )
        result, err = [each[name] for each in log]
        if result == states.FAIL:
            if err is not None:
                self.fail(str(err))
            else:
                self.fail('Screenshots were different.')
        elif result == states.ERROR:
            if err is not None:
                raise err
            raise exc.TestError() # todo
    finally:
        util.close_driver(driver)


class GossamerTestCase(unittest.TestCase): # pylint: disable=R0904
    """
    unittest case.
    """

    _skip_allowed = True
    _driver_ok = True
    _gossamer_test = None

    def setUp(self):
        super(GossamerTestCase, GossamerTestCase()).setUp()
        if not self._driver_ok:
            if self._skip_allowed:
                self.skipTest("Selenium Server is not available")
            else:
                raise exc.WebDriverConnectionFailed(
                    "Selenium Server is not available"
                )

    def runTest(self):
        """
        Replaced within :func:`run_gossamerfile`.
        """
        pass



"""
Application constants including run modes and exit codes.
"""

# Copyright (c) 2013 contributors; see AUTHORS.
# Licensed under the Apache License, Version 2.0
# https://www.apache.org/licenses/LICENSE-2.0

import os
import json


DEFAULT_TESTFILE = 'Gossamerfile'

DEFAULT_WEBDRIVER = 'http://localhost:4444/wd/hub'

LOCAL_WEBDRIVER_URL = os.environ.get('HUXLEY_WEBDRIVER_LOCAL', DEFAULT_WEBDRIVER)
REMOTE_WEBDRIVER_URL = os.environ.get('HUXLEY_WEBDRIVER_REMOTE', DEFAULT_WEBDRIVER)
ENV_DEFAULTS = json.loads(os.environ.get('HUXLEY_DEFAULTS', 'null'))

DEFAULT_DIFFCOLOR = '0,255,0'
DEFAULT_SCREENSIZE = '1024x768'
DEFAULT_BROWSER = 'firefox'

DATA_VERSION = 1

class _TestRunModes(object): # pylint: disable=R0903
    """
    Indicates what a given test run should be doing, e.g., recording.
    """
    RECORD = 1
    RERECORD = 2
    PLAYBACK = 3


class _ExitCodes(object): # pylint: disable=R0903
    """
    Exit codes...
    """
    OK = 0
    FAILED = 1
    ERROR = 2
    ARGUMENT_ERROR = 3 # CLI interface
    RECORDED_RUN_ERROR = 4 # Something went wrong with loading recorded records

class _OK(object): # pylint: disable=R0903,C0111
    def __unicode__(self): # pragma: no cover, pylint: disable=R0201
        return u'ok'
    def __str__(self): # pragma: no cover
        return self.__unicode__()

class _FAIL(object): # pylint: disable=R0903,C0111
    def __unicode__(self): # pragma: no cover, pylint: disable=R0201
        return u'FAIL'
    def __str__(self): # pragma: no cover
        return self.__unicode__()

class _ERROR(object): # pylint: disable=R0903,C0111
    def __unicode__(self): # pragma: no cover, pylint: disable=R0201
        return u'ERROR'
    def __str__(self): # pragma: no cover
        return self.__unicode__()

class _StatusCodes(object): # pylint: disable=R0903
    """
    Status codes, namely OK/FAIL/ERROR.
    """
    OK = _OK()
    FAIL = _FAIL()
    ERROR = _ERROR()


modes = _TestRunModes()
exits = _ExitCodes()
states = _StatusCodes()



"""
:func:`.dispatch` runs tests that have already been setup.

Tests can be setup through the command-line interface via
:func:`gossamer.cli.main` and through unittest via
:class:`gossamer.integration.GossamerTestCase`.
"""

# Copyright (c) 2013 contributors; see AUTHORS.
# Licensed under the Apache License, Version 2.0
# https://www.apache.org/licenses/LICENSE-2.0

from selenium.common.exceptions import WebDriverException

from gossamer.constant import modes, states
from gossamer import run, exc, util


def dispatch(driver, mode, tests, output=None, stop_on_error=False):
    """
    Given driver and a list of tests, dispatch the appropriate runs and
    return an exit code. For consumption by the CLI and unittest
    integration.
    """
    if not output:
        output = util.stdout_writer
    funcs = {
        modes.RECORD: (run.record, lambda x: (x.settings, )),
        modes.RERECORD: (run.rerecord, lambda x: (x.settings, x.steps)),
        modes.PLAYBACK: (run.playback, lambda x: (x.settings, x.steps))
    }
    result_log = {name: None for name, _ in tests.items()}
    error_log = {name: None for name, _ in tests.items()}
    try:
        for name, test in tests.items():
            result, err = funcs[mode][0](driver, *funcs[mode][1](test), output=output)
            if (not result or result in (states.FAIL, states.ERROR)) and stop_on_error:
                break
            result_log[name] = result
            error_log[name] = err
            if mode == modes.RECORD:
                util.write_recorded_run(test.settings.path, result)
    except exc.NoScreenshotsRecorded:
        raise
    except WebDriverException as exception:
        if 'not reachable' in exception.msg:
            raise exc.WebDriverWentAway('WebDriver cannot be reached.')
        elif 'Cannot find function' in exception.msg:
            raise exc.UnavailableBrowser(
                'WebDriver cannot seem to run the browser you selected. You '
                'may be missing a WebDriver dependency needed for that browser '
                '(e.g., chromium-chromedriver), or may need to use a separate '
                'WebDriver (e.g., InternetExplorerDriver).'
                )
        raise
    except Exception:
        raise
    return (result_log, error_log)



"""
Record, playback, &c a given test.
"""

# Copyright (c) 2013 contributors; see AUTHORS.
# Licensed under the Apache License, Version 2.0
# https://www.apache.org/licenses/LICENSE-2.0

import operator
import time

from selenium.common.exceptions import WebDriverException

from gossamer.constant import states, DATA_VERSION
from gossamer.step import Screenshot, Click, Key, Scroll, Text, Navigate
from gossamer.data import Point, Test
from gossamer import util, js, exc

__all__ = ['playback', 'record', 'rerecord', ]


def navigate(driver, url):
    """
    Navigate the driver to the given URL.
    """
    href, postdata = url
    driver.get('about:blank')
    driver.refresh()
    if not postdata:
        driver.get(href)
    else:
        driver.execute_script(js.get_post(href, postdata))
    _load_initial_js(driver)


def _load_initial_js(driver):
    """
    Split from :func:`.navigate` for calling within :func:`.record`
    after a URL change.
    """
    driver.execute_script(js.getGossamerEvents)
    driver.execute_script(js.pageChangingObserver)


def wait_until_loaded(driver):
    """
    Determine that a page has been loaded.
    """
    initial_timeout = 0
    while initial_timeout < 40:
        initial_timeout += 1
        if not driver.execute_script(js.isPageChanging(500)):
            break
        else:
            time.sleep(0.25)
    if initial_timeout == 40:
        raise exc.PlaybackTimeout(
            'Timed out while waiting for the initial load.'
        )


def _process_steps(steps, events, start_time):
    """
    Process events from the user agent into our objects.
    """
    for (timestamp, action, params) in events:
        if action == 'click':
            steps.append(
                Click(timestamp - start_time, Point(*params))
            )
        elif action == 'keyup':
            steps.append(
                Key(
                    timestamp - start_time,
                    key=params[0], shift=params[1],
                    eid=params[2][0], eid_val=params[2][1],
                    ecn=params[3][0], ecn_val=params[3][1],
                    ecl=params[4][0], ecl_val=params[4][1]
                )
            )
        elif action == 'scroll':
            steps.append(
                Scroll(timestamp - start_time, Point(*params))
            )

    # iterate through steps, and process 'raw' steps like Key into Text
    last_screenshot_time = None
    merges = []
    length = len(steps) - 1
    for i, step in enumerate(steps):
        if isinstance(step, Screenshot):
            last_screenshot_time = step.offset_time
        if isinstance(step, Key):
            if i != length and step.key != '\t' and isinstance(steps[i+1], Key):
                continue
            else:
                # tab keyup carries the element of the new field, so go back one
                step = step if step.key != '\t' else (steps[i-1] if i !=0 else None)
                if step is not None:
                    merges.append(
                        Text(
                            offset_time=step.offset_time,
                            value=step.value,
                            identifier_type=step.identifier_type,
                            identifier=step.identifier
                        )
                    )
        # not merging scrolls; might've been done for rendering side effects

    def _filter_steps(step):
        """
        Filter for whether a given step should go to playback.
        """
        return step.offset_time <= last_screenshot_time and \
            step.playback is True

    steps = sorted(
        filter(_filter_steps, steps + merges), # pylint: disable=W0141
        key=operator.attrgetter('offset_time')
    )
    return steps


def _begin_browsing(driver, settings):
    """
    Prepare the browser for the test to begin.
    """
    try:
        driver.delete_all_cookies()
        driver.set_window_size(*settings.screensize)
        navigate(driver, settings.navigate())
        if settings.cookies is not None and len(settings.cookies) > 0:
            for cookie in settings.cookies:
                driver.add_cookie(cookie)
            # selenium issue forces us to go to the domain in question
            # before setting a cookie for it -- so this assumes
            # we get only one domain in cookies.
            # https://code.google.com/p/selenium/issues/detail?id=1953
            driver.refresh()
            _load_initial_js(driver)
    except WebDriverException as exception:
        if exception.msg.startswith("Error communicating with the remote browser"):
            raise exc.WebDriverWentAway(
                "Cannot connect to the WebDriver browser: %s" % str(exception)
            )
        elif exception.msg.startswith("<unknown>: SecurityError: DOM Exception"):
            raise exc.WebDriverSecurityError(
                "WebDriver experienced a security error: %s" % str(exception)
            )
        raise


def _has_page_changed(url, driver_url):
    """
    Has the page's URL changed? Exclude everything after a hash.
    """
    return \
        url[0:(url.find('#') \
            if url.find('#') > 0 \
            else len(url))]\
        .rstrip('/') \
        != \
        driver_url[0:(driver_url.find('#') \
            if driver_url.find('#') > 0 \
            else len(driver_url))]\
        .rstrip('/')


def record(driver, settings, output):
    """
    Record a given test.
    """
    _begin_browsing(driver, settings)
    start_time = driver.execute_script(js.now)
    url = settings.url
    steps = []
    navs = []
    while True:
        if util.prompt("\nPress enter to take a screenshot, "
            "or type Q if you're done.", ('Q', 'q'), testname=settings.name):
            break
        # detect page changes
        if _has_page_changed(url, driver.current_url):
            navs.append(
                Navigate(
                    driver.execute_script(js.now) - start_time,
                    driver.current_url
                )
            )
            url = driver.current_url
            _load_initial_js(driver)
        output('Taking screenshot ... ', flush=True)
        screenshot_step = Screenshot(
            driver.execute_script(js.now) - start_time,
            len(steps) + 1
        )
        driver.save_screenshot(screenshot_step.get_path(settings))
        steps.append(screenshot_step)
        output(
            '%d screenshot%s in test.\n' % \
            (len(steps), 's' if len(steps) > 1 else '')
        )
    if len(steps) == 0:
        raise exc.NoScreenshotsRecorded(
            'No screenshots recorded for %s--please use at least one\n' % \
                settings.name
        )

    # now capture the events
    events = driver.execute_script('return window._getGossamerEvents();')
    if type(events) in (unicode, str) and events.startswith(
                                    'A script on this page may be busy, '
                                    'or it may have stopped responding.'):
        raise exc.TestError('Event-capturing script was unresponsive.')

    record = Test( # pylint: disable=W0621
        version = DATA_VERSION,
        settings = settings,
        steps = _process_steps(steps + navs, events, start_time)
    )

    util.prompt(
        "\n"
        "Up next, we'll re-run your actions to generate screenshots to "
        "ensure they \nare pixel-perfect when running automated. Press "
        "enter to start.", testname=settings.name
    )
    playback(driver, settings, record, output)

    return (record, None)


def rerecord(driver, settings, record, output): # pylint: disable=W0621
    """
    Rerecord a given test. :func:`.playback` handles it based on mode.
    """
    return playback(driver, settings, record, output)


def playback(driver, settings, record, output): # pylint: disable=W0621,R0912
    """
    Playback a given test.
    """
    if settings.desc:
        output("%s ... " % settings.desc, flush=True)
    else:
        output("Playing back %s ... " % settings.name, flush=True)

    _begin_browsing(driver, settings)
    wait_until_loaded(driver)

    state = states.OK
    err = None
    try:
        for step in record.steps:
            step.delayer(driver)
            timeout = 0
            while timeout < 40:
                timeout += 1
                if not driver.execute_script(js.isPageChanging(250)): # milliseconds
                    step.execute(driver, settings)
                    break
                else:
                    time.sleep(0.25)
            if timeout == 40:
                raise exc.PlaybackTimeout(
                    '%s timed out while waiting for the page to be static.' \
                        % settings.name
                )

    except Exception as exception: # pylint: disable=W0703
        if isinstance(exception, exc.ScreenshotsDiffer):
            state = states.FAIL
            err = exception
        else:
            state = states.ERROR
            if hasattr(exception, 'msg') and (exception.msg.startswith('element not visible') or
                exception.msg.startswith('Element is not currently visible')):
                err = exc.ElementNotVisible(
                    "Element was not visible when expected during playback. If "
                    "your playback depended on a significant rerender having been "
                    "done, then make sure you've waited until nothing is changing "
                    "before taking a screenshot."
                )
            else:
                err = exception

    output('%s' % str(state))
    if err:
        output(': %s' % str(err))
    return (state, err)




"""
Objects representing steps a user can take, e.g., make a click,
take a screenshot.
"""

# Copyright (c) 2013 contributors; see AUTHORS.
# Licensed under the Apache License, Version 2.0
# https://www.apache.org/licenses/LICENSE-2.0

import os
import time

from gossamer.constant import modes
from gossamer.exc import ScreenshotsDiffer
from gossamer.image import images_identical, image_diff, allowance
from gossamer import util


class TestStep(object): # pylint: disable=R0903
    """
    Base class of test actions, not useful in itself.
    """

    def __init__(self, offset_time):
        self.offset_time = offset_time

    def delayer(self, driver): # pylint: disable=R0201,W0613
        """
        Do not execute until...
        """
        time.sleep(0.25)

    def execute(self, driver, settings):
        """
        Called during :func:`.run.playback` or :func:`.run.rerecord`.
        """
        raise NotImplementedError

    def __json__(self):
        return {self.__class__.__name__: self.__dict__}


class Navigate(TestStep): # pylint: disable=R0903
    """
    Navigation to a new page.
    """
    playback = True

    def __init__(self, offset_time, url):
        super(Navigate, self).__init__(offset_time)
        self.url = url

    def execute(self, driver, settings):
        from gossamer.run import navigate, wait_until_loaded
        util.log.debug('Navigating to %s', self.url)
        navigate(driver, (self.url, None))
        wait_until_loaded(driver)



class Click(TestStep): # pylint: disable=R0903
    """
    Click action by the user.
    """

    playback = True

    def __init__(self, offset_time, pos):
        super(Click, self).__init__(offset_time)
        self.pos = pos

    def execute(self, driver, settings):
        util.log.debug("Clicking %s", self.pos)
        # Work around multiple bugs in WebDriver's implementation of click()
        driver.execute_script(
            'document.elementFromPoint(%d, %d).click();' % (self.pos.x, self.pos.y)
        )


class Key(TestStep): # pylint: disable=R0903,R0902,W0223
    """
    Typing action by the user.
    """

    playback = False

    def __init__(self, offset_time, key, shift=None,
        eid=None, ecn=None, ecl=None, eid_val=None, ecn_val=None, ecl_val=None
        ): # pylint: disable=R0913
        super(Key, self).__init__(offset_time)
        self.key = key
        self.shift = shift
        self.identifier = None
        self.identifier_type = None
        self.value = None

        # element.id
        self.eid = eid if eid and len(eid) > 0 else None
        # element.className
        self.ecn = ecn if ecn and len(ecn) > 0 else None
        # element.classList
        self.ecl = '.%s' % '. '.join(ecl) if ecl and len(ecl) > 0 else None

        self.eid_val = eid_val
        self.ecn_val = ecn_val
        self.ecl_val = ecl_val

        if not (self.eid or self.ecn or self.ecl):
            raise ValueError()
        if self.eid:
            self.identifier = self.eid
            self.identifier_type = 'id'
            self.value = self.eid_val
        elif self.ecl:
            self.identifier = self.ecl
            self.identifier_type = 'classlist'
            self.value = self.ecl_val
        elif self.ecn:
            self.identifier = self.ecn
            self.identifier_type = 'classname'
            self.value = self.ecn_val


class Text(TestStep): # pylint: disable=R0903
    """
    Text input, after being resolved from Keys.
    """

    playback = True

    def __init__(self, offset_time, identifier, identifier_type, value):
        super(Text, self).__init__(offset_time)
        self.identifier = identifier
        self.identifier_type = identifier_type
        self.value = value

    def delayer(self, driver):
        """
        Do not execute until...
        """
        time.sleep(1)

    def execute(self, driver, settings):
        util.log.debug(
            "Text '%s' into element '%s' by %s",
            self.value, self.identifier, self.identifier_type
        )
        funcs = {
            'id': driver.find_element_by_id,
            'classname': driver.find_element_by_class_name,
            'classlist': driver.find_element_by_css_selector
        }
        funcs[self.identifier_type](self.identifier).send_keys(self.value)


class Screenshot(TestStep):
    """
    Screenshot taken by the user.
    """

    playback = True

    def __init__(self, offset_time, num):
        super(Screenshot, self).__init__(offset_time)
        self.num = num

    def delayer(self, driver):
        """
        Do not execute until...
        """
        time.sleep(1)

    def get_path(self, settings):
        """
        Path to screenshot.
        """
        return os.path.join(settings.path, 'screenshot' + str(self.num) + '.png')

    def execute(self, driver, settings):
        util.log.debug("Taking screenshot %s", self.num)
        original = self.get_path(settings)
        new = os.path.join(settings.path, 'last', 'screenshot%s.png' % self.num)
        if settings.mode == modes.RERECORD:
            driver.save_screenshot(original)
        else:
            driver.save_screenshot(new)
            if not images_identical(original, new, allowance(settings.browser)):
                if settings.save_diff:
                    diffpath = os.path.join(settings.path, 'diff.png')
                    diff = image_diff(original, new, diffpath, settings.diffcolor)
                    raise ScreenshotsDiffer(
                        'Screenshot %s was different; compare %s with %s. See %s '
                        'for the comparison. diff=%r' % (
                            self.num, original, new, diffpath, diff
                        )
                    )
                else:
                    raise ScreenshotsDiffer('Screenshot %s was different.' % self.num)


class Scroll(TestStep): # pylint: disable=R0903
    """
    Scrolling action on the page.
    """

    playback = True

    def __init__(self, offset_time, pos):
        super(Scroll, self).__init__(offset_time)
        self.pos = pos

    def execute(self, driver, settings):
        util.log.debug("Scrolling to %s", self.pos)
        driver.execute_script("window.scrollBy(%s, %s)" % (self.pos.x, self.pos.y))




"""
setup
"""

from setuptools import setup
import os, re

with open('gossamer/__init__.py') as version:
    VERSION = re.compile(r".*__version__ = '(.*?)'", re.S).match(version.read()).group(1)

# rst, written by `make doc`
README = ''
if os.path.exists('README'):
    with open('README', 'r') as readme:
        README = readme.read()

setup(
    name = 'gossamerui',
    provides = 'gossamer',
    version = VERSION,
    packages=['gossamer', ],
    install_requires = [
        'selenium >= 2.35.0',
        'plac == 0.9.1',
        'pillow >= 2.1.0'
    ],
    classifiers=[
      'Development Status :: 4 - Beta',
      'Environment :: Console',
      'Intended Audience :: Developers',
      'Operating System :: OS Independent',
      'License :: OSI Approved :: Apache Software License',
      'Programming Language :: Python',
      'Programming Language :: Python :: 2',
      'Topic :: Software Development :: Testing',
      'Topic :: Software Development :: Quality Assurance',
      'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    package_data = {
      'gossamer': [
        '*.js',
      ],
    },
    entry_points = {
        'console_scripts': [
            'gossamer = gossamer.cli:main'
        ],
        'setuptools.installation': [
            'eggsecutable = gossamer.cli:main',
        ]
    },
    license = 'Apache 2.0',
    keywords = 'selenium webdriver testing regression ui automated visual diff screenshot huxley',
    description = 'Website user interface regression testing',
    long_description = README,
    url = 'https://github.com/ijl/gossamer',
    download_url = 'https://github.com/ijl/gossamer/archive/%s.tar.gz' % VERSION,
    author = 'Jack Lutz',
    author_email = 'uijllji@gmail.com',
    zip_safe=True
)



"""
Utilities.
"""

# Copyright (c) 2013 contributors; see AUTHORS.
# Licensed under the Apache License, Version 2.0
# https://www.apache.org/licenses/LICENSE-2.0

import os
import sys
import json
import operator
import ConfigParser
from selenium.common.exceptions import WebDriverException
import urllib2

from selenium import webdriver  # pylint: disable=F0401

from gossamer import exc

from gossamer.constant import modes,  DEFAULT_DIFFCOLOR, \
    REMOTE_WEBDRIVER_URL, DATA_VERSION

def logger(name, level=None):
    """
    Create logger with appropriate level.
    """
    import logging
    ret = logging.getLogger(name)
    ret.addHandler(logging.StreamHandler())
    ret.setLevel(getattr(logging, level) if level else logging.INFO)
    return ret

# level can be overriden to DEBUG in CLI with -v
log = logger(__name__, 'INFO')


def stdout_writer(content=None, flush=False):
    """
    Write to stdout. Pluggable in :func:`main.dispatch` for suppressing
    output.
    """
    if content:
        sys.stdout.write(content)
    if flush:
        sys.stdout.flush()


def null_writer(content=None, flush=False): # pylint: disable=W0613
    """
    No output written.
    """
    pass


DRIVERS = {
    'firefox': webdriver.Firefox,
    'chrome': webdriver.Chrome,
    'ie': webdriver.Ie,
    'opera': webdriver.Opera
}

CAPABILITIES = {
    'firefox': webdriver.DesiredCapabilities.FIREFOX,
    'chrome': webdriver.DesiredCapabilities.CHROME,
    'ie': webdriver.DesiredCapabilities.INTERNETEXPLORER,
    'opera': webdriver.DesiredCapabilities.OPERA
}


class Encoder(json.JSONEncoder):
    """
    Overriden JSON encoder for calling __json__.
    """

    def default(self, obj): # pylint: disable=E0202
        """
        Test and steps implement a __json__ attribute.
        """
        if hasattr(obj, '__json__'):
            return obj.__json__()
        return json.JSONEncoder.default(self, obj)


def _import_run_v1(rec):
    """
    Deserialize for Test.version == 1
    """
    from gossamer.data import Test, Settings, Point
    from gossamer import step
    mode_trans = {1: modes.RECORD, 2: modes.RERECORD, 3: modes.PLAYBACK}
    steps = []
    for each in rec['steps']:
        key, val = each.items()[0]
        if 'pos' in val:
            val['pos'] = Point(**val['pos'])
        steps.append(getattr(step, key)(**val))
    for key, val in rec['settings'].items():
        if key == 'mode':
            rec['settings'][key] = mode_trans[val]
    steps = sorted(steps, key=operator.attrgetter('offset_time'))
    test = Test(
        version=rec['version'],
        settings=Settings(**rec['settings']),
        steps=steps
    )
    return test


def import_recorded_run(inc):
    """
    Given a JSON object, create our objects.
    """
    if len(inc) > 1:
        raise ValueError()
    for _, rec in inc.items(): # 1-element list
        obj = rec
    if obj['version'] == 1:
        return _import_run_v1(obj)
    else:
        raise NotImplementedError()


def read_recorded_run(filename):
    """
    Load a serialized run.
    """
    try:
        if os.path.getsize(filename) <= 0:
            raise exc.RecordedRunEmpty('%s is empty' % filename)
    except OSError:
        raise exc.RecordedRunDoesNotExist('%s does not exist' % filename)
    with open(filename, 'r') as fp:
        try:
            rec = json.loads(fp.read())
        except ValueError: # couldn't parse
            raise exc.CouldNotParseRecordedRun('Could not parse %s' % filename)
    return import_recorded_run(rec)


def write_recorded_run(filename, output):
    """
    Serialize a recorded run to a JSON file.
    """
    from gossamer.data import Test
    if not isinstance(output, Test):
        raise ValueError()
    try:
        with open(os.path.join(filename, 'record.json'), 'w') as fp:
            fp.write(json.dumps(output, cls=Encoder))
    except Exception as exception: # todo
        raise exception
    return True


def check_driver(url):
    """
    For a Selenium Server URL, see if the server appears to be up.
    """
    try:
        code = urllib2.urlopen(url).getcode()
    except Exception: # pylint: disable=W0703
        code = 500
    return code in (200, 301, 302)


def get_driver(browser, local_webdriver=None, remote_webdriver=None):
    """
    Get a webdriver. The caller is responsible for closing the driver.

    Browser is required. Local and remote are optional, with remote
    taking precedence.
    """
    if local_webdriver and not remote_webdriver:
        driver_url = local_webdriver
    else:
        driver_url = remote_webdriver or REMOTE_WEBDRIVER_URL
    try:
        driver = webdriver.Remote(driver_url, CAPABILITIES[browser])
    except KeyError:
        raise exc.InvalidBrowser(
            'Invalid browser %r; valid browsers are %r.' % (browser, DRIVERS.keys())
        )
    except urllib2.URLError as exception:
        raise exc.WebDriverConnectionFailed(
            'We cannot connect to the WebDriver %s -- is it running?' % driver_url
        )
    except WebDriverException as exception:
        if exception.msg.startswith('The path to the driver executable must be set'):
            raise exc.InvalidWebDriverConfiguration(
                'WebDriver cannot locate the driver for %s: %s' % (browser, exception.msg)
            )
        raise
    return driver


def close_driver(driver):
    """
    Close the driver, or fail silently if it doesn't exist.
    """
    try:
        driver.quit()
    except UnboundLocalError: # pragma: no cover
        pass
    except AttributeError: # pragma: no cover
        pass
    return True

def prompt(display, options=None, testname=None):
    """
    Given text as `display` and optionally `options` as an
    iterable containing acceptable input, returns a boolean
    of whether the prompt was met.
    """
    sys.stdout.write(display)
    sys.stdout.write('\n')
    sys.stdout.flush()
    inp = raw_input('gossamer%s >>> ' % (':'+testname if testname else ''))
    if options:
        if inp in options:
            return True
        return False
    return True


def _postdata(arg):
    """
    Given CLI input `arg`, resolve postdata.
    """
    if arg:
        if arg == '-':
            return sys.stdin.read()
        else:
            with open(arg, 'r') as f:
                data = json.loads(f.read())
            return data
    return None


def verify_and_prepare_files(filename, testname, mode, overwrite):
    """
    Prepare directories on file system, including clearing existing data
    if necessary.
    """
    log.debug('%s: %s', testname, filename)
    if os.path.exists(filename):
        if mode == modes.RECORD: # todo weirdness with rerecord
            if os.path.getsize(filename) > 0:
                if not overwrite and not prompt(
                    '\n%s already exists--clear existing '
                    'screenshots and overwrite test? [Y/n] ' \
                        % testname,
                    ('Y', 'y')
                    ):
                    raise exc.DoNotOverwrite(
                        "Aborting because we don't wish to overwrite %s." % \
                        testname
                    )
                for each in os.listdir(filename):
                    if each.split('.')[-1] in ('png', 'json'):
                        os.remove(os.path.join(filename, each))
                for each in os.listdir(os.path.join(filename, 'last')):
                    if each.split('.')[-1] == 'png':
                        os.remove(os.path.join(filename, 'last', each))
    else:
        if mode == modes.RECORD:
            os.makedirs(filename)
            os.makedirs(os.path.join(filename, 'last'))
            # todo exc
        else:
            raise exc.EmptyPath('%s does not exist\n' % filename)
    return True


def make_tests(test_files, mode, data_dir, rewrite_url=None, **kwargs): # pylint: disable=R0914,R0912,R0915
    """
    Given a list of gossamer test files, a mode, working directory, and
    options as found on the CLI interface, make tests for use by the
    dispatcher.
    """
    from gossamer.data import Settings, Test

    postdata = _postdata(kwargs.pop('postdata', {}))
    diffcolor = tuple(
        int(x) for x in (kwargs.pop('diffcolor', None) or DEFAULT_DIFFCOLOR).split(',')
    )

    tests = {}
    names = kwargs.pop('names', None)
    overwrite = kwargs.get('overwrite', False)
    existing_names = []

    for file_name in test_files:

        config = ConfigParser.SafeConfigParser(
            allow_no_value=True
        )
        config.read([file_name])

        for testname in config.sections():

            existing_names.append(testname)
            if names and (testname not in names):
                continue
            if testname in tests:
                raise exc.DuplicateTestName('Duplicate test name %s' % testname)

            test_config = dict(config.items(testname))
            filename = os.path.join(data_dir, testname)

            if mode != modes.PLAYBACK:

                url = test_config.get('url', None)
                if not url:
                    raise exc.InvalidGossamerfile(
                        '%s did not have a `url` argument' % testname
                    )
                url = url.strip("'").strip('"')

                cookies = test_config.get('cookies', None)
                if cookies and len(cookies) > 0:
                    cookies = json.loads(cookies)
                else:
                    cookies = None

                screensize = tuple(
                    int(x) for x in
                        (kwargs.pop('screensize', None) or \
                        test_config.get('screensize', '1024x768')
                ).split('x'))

                sa_browser = test_config.get('browser', None)
                kw_browser = kwargs.get('browser', None)
                if sa_browser and kw_browser and sa_browser != kw_browser:
                    raise exc.DifferentBrowser(
                        "Different browser given in command-line than "
                        "is on recorded run. Screenshots may not match, "
                        "so aborting. Please re-record."
                    )
                browser = kw_browser if kw_browser is not None else sa_browser

                recorded_run = None
                settings = Settings(
                    name=testname,
                    desc=test_config.get('desc', None),
                    url=url,
                    mode=mode,
                    path=filename,
                    browser=browser,
                    screensize=screensize,
                    postdata=postdata or test_config.get('postdata'),
                    diffcolor=diffcolor,
                    save_diff=kwargs.pop('save_diff', None),
                    cookies=cookies
                )

            else:
                recorded_run = read_recorded_run(os.path.join(filename, 'record.json'))
                settings = recorded_run.settings
                settings.path = filename
                if rewrite_url:
                    settings.url = rewrite_url(settings.url)

            verify_and_prepare_files(filename, testname, mode, overwrite)

            # load recorded runs if appropriate
            if mode != modes.RECORD:
                recorded_run = read_recorded_run(os.path.join(filename, 'record.json'))
            else:
                recorded_run = None

            tests[testname] = Test(version=DATA_VERSION, settings=settings, steps=recorded_run)

    if names:
        for name in names:
            if name not in existing_names:
                raise exc.UnknownTestName(
                    "'%s' is not in the parsed files. Valid options are: %r" % \
                    (name, existing_names)
                )

    return tests




"""
JavaScript to be executed in the testing user agent.
"""

# Copyright (c) 2013 contributors; see AUTHORS.
# Licensed under the Apache License, Version 2.0
# https://www.apache.org/licenses/LICENSE-2.0

import pkg_resources
import json


def _get_javascript(name):
    """
    Convenience for reading JavaScript from package.
    """
    return pkg_resources.resource_stream('gossamer', '%s.js' % name).read()


# https://developer.mozilla.org/en-US/docs/Web/API/MutationObserver
pageChangingObserver = _get_javascript('pageChangingObserver')


getGossamerEvents = _get_javascript('getGossamerEvents')


def isPageChanging(timeout): # pragma: no cover
    """
    Has page changed within the given `timeout`, in milliseconds, and are
    there no XMLHTTP requests active?
    """
    return """
return window._gossamerIsPageChanging(%s);
""" % timeout


def get_post(url, postdata): # pragma: no cover
    """
    Retrieve data for navigate.
    """
    markup = '\n'.join([
        '<form method="post" action="%s">' % url,
        '\n'.join(['<input type="hidden" name="%s" />' % k for k in postdata.keys()]),
        '</form>'
    ])

    script = 'var container = document.createElement("div"); container.innerHTML = %s;' \
            % json.dumps(markup)

    for (i, v) in enumerate(postdata.values()):
        if not isinstance(v, basestring):
            v = json.dumps(v)
        script += 'container.children[0].children[%d].value = %s;' % (i, json.dumps(v))

    script += 'document.body.appendChild(container);'
    script += 'container.children[0].submit();'
    return '(function(){ ' + script + '; })();'


now = """
return Date.now();
"""
